/*	Copyright (C) 2021 Mesharsky
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, see <http://www.gnu.org/licenses/>.

	// Credits n.o.x for his old base plugin.
	// Credits Bioly for helping.
*/

#pragma semicolon 1

// Includes
#include <sourcemod>
#include <multicolors>
#include <sdktools>
#include <discord>

#pragma newdecls required

// SQL Queries to Create the Database Table and to Insert the data into SQL Table.
static const char mReport_CreateSQL[] = "CREATE TABLE IF NOT EXISTS `mReport` (`id` INT(20) NOT NULL AUTO_INCREMENT, `client_steamid` VARCHAR(64) NOT NULL, `client_name` VARCHAR(128) NOT NULL, `target_steamid` VARCHAR(64) NOT NULL, `target_name` VARCHAR(128) NOT NULL, `server_name` VARCHAR(256) NOT NULL, `report_reason` VARCHAR(256) NOT NULL, `game` VARCHAR(256) NOT NULL, `time` INT(20) NOT NULL, PRIMARY KEY (id)) ENGINE=MyISAM DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_general_ci";
static const char mReport_InsertSQL[] = "INSERT INTO `mReport` (client_steamid, client_name, target_steamid, target_name, server_name, report_reason, game, time) VALUES ('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%d')";

#define CHAT_PREFIX "\x05[ mReport ]\x01 " // You can change it to your own. Make sure you know the colors.
#define GAME "csgo" // Don`t touch it, unless you rewrite it for CS:S. Its needed for game-type in database table.


// Bool to indicate whenever player choose option to report with custom reason.
bool g_mInputReason[MAXPLAYERS + 1];

// Timer between reporting players.
int g_mCommandWaitTime[MAXPLAYERS + 1];

// Integer to indicate the target that will be reported.
int g_iBanTarget[MAXPLAYERS + 1];

// DV
Database db = null;

// Convars
ConVar g_cvServerName;
ConVar g_cvAnnounceAdmin;
ConVar g_cvCommandTimeInterval;

ConVar g_cvDiscordIntegration;
ConVar g_cvDiscord_WebHook;
ConVar g_cvDiscord_HookBotBame;

public Plugin myinfo = 
{
	name = "[CSGO] Advanced Report System", 
	author = "Mesharsky", 
	description = "SQL + Discord Report System for CSGO Servers", 
	version = "4.0.1", 
	url = "https://go-code.pl"
};

public void OnPluginStart()
{
	// Commands to open report menu.
	RegConsoleCmd("sm_report", Command_Report, "Report Menu");
	
	// Needed for custom reason in chat.
	AddCommandListener(ChatListener, "say");
	AddCommandListener(ChatListener, "say2");
	AddCommandListener(ChatListener, "say_team");
	
	// Basic plugin cvars
	g_cvAnnounceAdmin = CreateConVar("mAnnounceAdmin", "1", "Whether to notify the admin if he is on the server about the notification?");
	g_cvServerName = CreateConVar("mServerName", "Only Mirage", "Enter the name of your server as it should appear in the database");
	g_cvCommandTimeInterval = CreateConVar("mCommandTimeInterval", "30", "Enter the delay ");
	
	// Discord Integration.
	g_cvDiscordIntegration = CreateConVar("mDiscordIntegration", "0", "[ Discord ] Enable or disable integration with Discord?", FCVAR_NONE, true, 0.0, true, 1.0);
	g_cvDiscord_WebHook = CreateConVar("mDiscordWebHook", "", "[ Discord ] Webhook for a Discord channel");
	g_cvDiscord_HookBotBame = CreateConVar("mDiscordHookBotBame", "mReport", "[ Discord ] Bot name (author of the Discord message).");
	
	// Try to connect to database straight on PluginStart
	DB_StartConnection();
}

public void OnClientPutInServer(int client)
{
	// Needs to be on false, so our chat will work. If it's set to true, chat will be broken on the server
	g_mInputReason[client] = false;
	
	// We have to set the Wait value to 0 so player will be able to report someone.
	g_mCommandWaitTime[client] = 0;
}

public void OnMapStart()
{
	// Config file for cvars.
	AutoExecConfig(true, "mReport_Config");
	
	// Download and precache notification admin sound.
	AddFileToDownloadsTable("sound/mReport/mNotification_admin.mp3");
	PrecacheSound("*/mReport/mNotification_admin.mp3", true);
}

public void DB_StartConnection()
{
	char SQL_Error[255];
	DBResultSet SQL_Handler;
	
	if (SQL_CheckConfig("mReport")) // Check if the configuration file is found in databases.cfg
	{
		db = SQL_Connect("mReport", true, SQL_Error, 255);
		
		if (db == null)
			SetFailState("[ mReport - Startup Error ] There is some problem with connecting to the database. Please check if you have right database credentials in: addons/sourcemod/configs/databases.cfg.");
	}
	else
		SetFailState("[ mReport - Startup Error ] Please fill up database credentials in: csgo/sourcemod/configs/databses.cfg. Key value for database tree is: mReport");
	
	if ((SQL_Handler = SQL_Query(db, mReport_CreateSQL)) == null) // If everything goes okay then start the connection.
	{
		SQL_GetError(db, SQL_Error, sizeof(SQL_Error));
		
		LogError("[ mReport - Error] There is some error when creating database: %s", SQL_Error);
		LogError("[ mReport - Error] Query that has been used: %s", mReport_CreateSQL);
	}
	
	delete SQL_Handler;
}

// Main menu to report players.
public Action Command_Report(int client, int args)
{
	// We need to count all the people on the server, so they will be listed in our menu.
	int count = 0;
	
	Menu menu = new Menu(MenuReport_Handler, MENU_ACTIONS_ALL);
	
	menu.SetTitle("[ mReport ] Select the player you want to report");
	
	// Loop for catching all the players on the server.
	for (int i = 1; i < MaxClients; i++)
	{
		if (!IsValidClient(i))
			continue;
		
		char sInfo[32];
		char sName[32];
		
		Format(sInfo, sizeof(sInfo), "%d", i);
		GetClientName(i, sName, sizeof(sName));
		
		menu.AddItem(sInfo, sName);
		
		count++;
	}
	
	// If can't find anyone, show this message in the menu
	if (count == 0)
		menu.AddItem("NULL", "[ n-o-n-e ] Somehow there is no player on the server that you can report back");
	
	menu.Display(client, MENU_TIME_FOREVER);
	
	return Plugin_Handled;
}

public int MenuReport_Handler(Menu menu, MenuAction action, int param1, int param2)
{
	switch (action)
	{
		case MenuAction_Select:
		{
			char sItem[32];
			menu.GetItem(param2, sItem, sizeof(sItem));
			
			mReport_Reasons(param1, StringToInt(sItem));
		}
		case MenuAction_DrawItem:
		{
			char info[32];
			menu.GetItem(param2, info, sizeof(info));
			if (StrEqual(info, "NULL"))
				return ITEMDRAW_DISABLED;
		}
		case MenuAction_End:
		delete menu;
	}
	return 0;
}

public void mReport_Reasons(int client, int target)
{
	// Store target for menu handler
	if (target < 1 || target > MaxClients || !IsClientConnected(target))
	{
		ReplyToCommand(client, "Invalid player index for the report");
		return;
	}
	
	g_iBanTarget[client] = GetClientUserId(target);
	
	Menu menu = new Menu(MenuReport_SubHandler, MENU_ACTIONS_ALL);
	menu.SetTitle("[ mReport ] Select the reason for reporting the suspect ➫ [ %N ]", target);
	
	// On the left: What will be sent into database
	// On the right: What will be visible in the menu for player
	// WARNING: DO NOT TOUCH LEFT SIDE of mCustomReason!!!
	menu.AddItem("Cheating", "Cheating");
	menu.AddItem("Abuse of the microphone", "Abuse of the microphone");
	menu.AddItem("Deliberate interference with gameplay", "Deliberate interference with gameplay");
	menu.AddItem("Insulting", "Insulting");
	menu.AddItem("Camping", "Camping");
	
	menu.AddItem("mCustomReason", "Enter your own reason");
	
	menu.Display(client, 0);
}

public int MenuReport_SubHandler(Menu menu, MenuAction action, int param1, int param2)
{
	switch (action)
	{
		case MenuAction_Select:
		{
			int style;
			char info[64], display[128];
			menu.GetItem(param2, info, sizeof(info), style, display, sizeof(display));
			
			if (StrEqual(info, "mCustomReason"))
			{
				CPrintToChat(param1, " %s Enter your reason for reporting in the chat...", CHAT_PREFIX);
				g_mInputReason[param1] = true;
			}
			else
			{
				int target = GetClientOfUserId(g_iBanTarget[param1]);
				if (target != 0)
					mReport_SendData(param1, target, info);
				else
					LogError("The report has failed because the target has already quit or is invalid.");
			}
		}
		
		case MenuAction_DrawItem:
		{
			char info[32];
			menu.GetItem(param2, info, sizeof(info));
			if (StrEqual(info, "NULL"))
				return ITEMDRAW_DISABLED;
			else if (StrEqual(info, "-TARGET-"))
				return ITEMDRAW_IGNORE;
		}
		
		case MenuAction_End:
		delete menu;
	}
	return 0;
}

public Action ChatListener(int client, const char[] command, int args)
{
	if (!IsValidClient(client))
		return Plugin_Handled;
	
	if (g_mInputReason[client])
	{
		char sMessage[MAX_MESSAGE_LENGTH];
		
		GetCmdArgString(sMessage, sizeof(sMessage));
		StripQuotes(sMessage);
		
		g_mInputReason[client] = false;
		
		int target = GetClientOfUserId(g_iBanTarget[client]);
		if (target != 0)
			mReport_SendData(client, target, sMessage);
		else
		{
			LogError("The report has failed because the target has already quit or is invalid.");
			return Plugin_Handled;
		}	
		
		return Plugin_Handled;
	}
	
	return Plugin_Continue;
}

public void mReport_SendData(int client, int target, const char[] s_mReason)
{
	char client_authid[64];
	char client_name[64];
	char target_name[64];
	char target_authid[64];
	char server_name[64];
	char Discord_WebHook[512];
	char Discord_Message[512];
	char Discord_HookBotBame[64];
	
	char escape_client_name[MAX_NAME_LENGTH * 2 + 1];
	char escape_target_name[MAX_NAME_LENGTH * 2 + 1];
	
	GetClientName(client, client_name, MAX_NAME_LENGTH);
	GetClientName(target, target_name, sizeof(target_name));
	GetClientAuthId(client, AuthId_SteamID64, client_authid, sizeof(client_authid));
	GetClientAuthId(target, AuthId_SteamID64, target_authid, sizeof(target_authid));
	
	g_cvServerName.GetString(server_name, sizeof(server_name));
	g_cvDiscord_WebHook.GetString(Discord_WebHook, sizeof(Discord_WebHook));
	g_cvDiscord_HookBotBame.GetString(Discord_HookBotBame, sizeof(Discord_HookBotBame));
	
	if (g_mCommandWaitTime[client] == 0 || g_mCommandWaitTime[client] <= GetTime())
	{
		if (g_cvDiscordIntegration.BoolValue)
		{
			Format(Discord_Message, sizeof(Discord_Message), "[ %s ] Player %s ( %s ), reports %s ( %s ), reason of the report: %s", server_name, client_name, client_authid, target_name, target_authid, s_mReason);
			
			DiscordWebHook hook = new DiscordWebHook(Discord_WebHook);
			hook.SetUsername(Discord_HookBotBame);
			hook.SlackMode = false;
			hook.SetContent(Discord_Message);
			hook.Send();
			
			delete hook;
		}
		
		db.Escape(client_name, escape_client_name, sizeof(escape_client_name));
		db.Escape(target_name, escape_target_name, sizeof(escape_target_name));
		
		char query[1024];
		Format(query, sizeof(query), mReport_InsertSQL, client_authid, escape_client_name, target_authid, escape_target_name, server_name, s_mReason, GAME, GetTime());
		
		db.SetCharset("utf8mb4");
		db.Query(DB_CallBack, query);
		
		CPrintToChat(client, " %s Player reported successfully {darkred}%N", CHAT_PREFIX, target);
		
		if (g_cvAnnounceAdmin.BoolValue)
		{
			for (int i = 1; i < MaxClients; i++)
			{
				if (IsValidClient(i))
				{
					if (GetAdminFlag(GetUserAdmin(i), Admin_Generic) || GetAdminFlag(GetUserAdmin(i), Admin_Root))
					{
						CPrintToChat(i, "{yellow}--- {lightred}REPORT NOTIFICATION {yellow}---");
						CPrintToChat(i, "Report from the player: {green}%s", client_name);
						CPrintToChat(i, "The report concerns the player: {darkred}%s", target_name);
						CPrintToChat(i, "Reason for the report: {purple}%s", s_mReason);
						CPrintToChat(i, "{green}A copy of the report was also generated in the console");
						CPrintToChat(i, "{yellow}--- {lightred}REPORT NOTIFICATION {yellow}---");
						
						PrintToConsole(i, "--- REPORT NOTIFICATION ---");
						PrintToConsole(i, "Report from the player: %s", client_name);
						PrintToConsole(i, "The report concerns the player: %s", target_name);
						PrintToConsole(i, "Reason for the report: %s", s_mReason);
						PrintToConsole(i, "--- REPORT NOTIFICATION ---");
						
						EmitSoundToClient(i, "*/mReport/mNotification_admin.mp3");
					}
				}
			}
		}
		
		g_mCommandWaitTime[client] = GetTime() + g_cvCommandTimeInterval.IntValue;
	}
	else
	{
		CPrintToChat(client, " %s {yellow}You have already submitted a player, wait {green}%d {yellow}seconds and try again.", CHAT_PREFIX, g_mCommandWaitTime[client] - GetTime());
		return;
	}
}

public int DB_CallBack(Handle owner, Handle query, const char[] error, any data)
{
	if (query == null)
	{
		LogError("[ mReport ] Information could not be saved (error: %s)", error);
		return;
	}
}

stock bool IsValidClient(int client)
{
	if (client <= 0)return false;
	if (client > MaxClients)return false;
	if (!IsClientConnected(client))return false;
	if (IsClientReplay(client))return false;
	if (IsFakeClient(client))return false;
	if (IsClientSourceTV(client))return false;
	return IsClientInGame(client);
}
